@@ -1,9 +1,10 @@
-from pyprotobuf.visitor import Visitor
+from pyprotobuf.codegenerator import CodeGenerator
 
-class CommentResolver(Visitor):
+class CommentResolver(CodeGenerator):
     ''' associate comments with suceeding nodes 
     '''
-    def __init__(self):
+    def __init__(self, frame=None):
+        super(CommentResolver, self).__init__(frame)
         self.last_comment = None
 
     def visit(self, parent):
