@@ -39,7 +39,6 @@ TOKEN_ERRORTOKEN = 52
 TOKEN_N_TOKENS = 53
 TOKEN_NT_OFFSET = 256
 
-
 TOKEN_LBRACE = 26
 TOKEN_RBRACE = 27
 TOKEN_AT = 50
@@ -65,12 +64,22 @@ TOKEN_SEMI = 13
 
 import re
 import logging
+
 logger = logging.getLogger('pyprotobuf.parser')
+
+
 def identifier(scanner, token): return TOKEN_NAME, token
+
+
 def operator(scanner, token):   return "OPERATOR", token
+
+
 def digit(scanner, token):      return "DIGIT", token
+
+
 def end_stmnt(scanner, token):  return "END_STATEMENT"
- 
+
+
 scanner = re.Scanner([
     (r"[a-zA-Z_][\.\w]*", lambda scanner, token: (TOKEN_NAME, token)),
     (r"//.*", lambda scanner, token: (TOKEN_COMMENT, token[2:])),
@@ -80,31 +89,44 @@ scanner = re.Scanner([
     (r"[\t ]+", lambda scanner, token: (TOKEN_SPACE, token)),
     (r'"(?:(?:\\.)|(?:[^"]))*"', lambda scanner, token: (TOKEN_STRING, token)),
     (r"\n", lambda scanner, token: (TOKEN_NL, token)),
-    ])
- 
+])
+
+
 class ParseError(Exception):
     def __repr__(self):
         return 'Parse Error: %s [%s:%s]' % (self.msg, self.line, self.col)
-        
+
+
 class ProtoParser(object):
+    """ Parses a single proto file to an AST.
+    """
+
     def __init__(self):
         pass
-    
+
     def parse_string(self, s):
         ''' proto =( message | extend | enum | import | package | option | ';' ) ; 
         '''
         tokenizer = Tokenizer(s)
-        self.protonode = ProtoNode()
+
+        # the root is an anonymous package node
+        self.root = Package()
+
+        self.protonode = FileNode()
+
+        self.root.add_child(self.protonode)
+
         while tokenizer.has_next():
-            token =  tokenizer.next_token()
+            token = tokenizer.next_token()
             if self.check_single_line_comment(token):
                 self.protonode.add_child(self.read_single_line_comment(tokenizer))
             elif self.check_multiline_line_comment(tokenizer, token):
                 self.protonode.add_child(self.read_multi_line_comment(tokenizer))
             if token.type == TOKEN_NAME:
                 self.protonode.add_child(self.parse_toplevel(tokenizer, token))
-        return self.protonode
-        
+
+        return self.root
+
     def parse_toplevel(self, tokenizer, token):
         methods = {
             'package': self.parse_package_stmt,
@@ -117,38 +139,38 @@ class ProtoParser(object):
             'service': self.parse_service_stmt
         }
         method = methods.get(token.value, None)
-        
+
         logger.debug('Parse top-level: %s', token.value)
-        
+
         if method is None:
             raise Exception("Unknown toplevel token %s" % token)
-            
+
         node = method(tokenizer)
         return node
-    
+
     def check_names(self, token, names):
         # TODO: varargs
         return token.type == TOKEN_NAME and token.value in names
 
     def check_single_line_comment(self, current_token):
         return current_token.type == TOKEN_OP and current_token.value == '//'
-        
+
     def check_multiline_line_comment(self, t, current_token):
-        return current_token.type == TOKEN_OP and current_token.value == '/' and t.readahead().value =='*'
-    
+        return current_token.type == TOKEN_OP and current_token.value == '/' and t.readahead().value == '*'
+
     def read_single_line_comment(self, t):
         comment = []
         while t.has_next():
-            token =  t.next_token(False)
+            token = t.next_token(False)
             if token.value == "\n":
                 t.line += 1
                 break
             comment.append(token.value)
-        logger.debug('Read comment %s',  ' '.join(comment))            
+        logger.debug('Read comment %s', ' '.join(comment))
         node = CommentNode()
         node.value = ' '.join(comment)
         return node
-        
+
     def read_multi_line_comment(self, t):
         star = t.next_token()
         assert star.value == '*'
@@ -157,96 +179,96 @@ class ProtoParser(object):
             token = t.next_token()
             if token.value == "\n":
                 t.line += 1
-                
+
             if token.value == '*' and t.readahead().value == '/':
                 slash = t.next_token()
                 assert slash.value == '/'
                 break
-                
+
             comment.append(token.value)
-        logger.debug('Read comment %s',  '\n'.join(comment))            
+        logger.debug('Read comment %s', '\n'.join(comment))
         node = CommentNode()
         node.value = '\n'.join(comment)
         return node
-        
+
     def read_name(self, t):
-        return t.next_token().value 
-               
+        return t.next_token().value
+
     def read_dotted(self, t):
         tok = t.next_token()
         while t.has_next():
-            more =  t.readahead()
-            if more.value != '.' :
+            more = t.readahead()
+            if more.value != '.':
                 break
             t.next_token()
             tok.value = tok.value + '.' + self.read_dotted(t).value
-        return tok       
-        
+        return tok
+
     def read_string_or_id(self, t):
         tok = t.next_token()
         assert tok.type == TOKEN_STRING or tok.type == TOKEN_NAME, "Expected ident/string literal. Got %s" % tok
         if tok.type == TOKEN_STRING:
-            tok.value = tok.value.strip("'"+'"')
+            tok.value = tok.value.strip("'" + '"')
         return tok
-        
+
     def read_string_or_id_or_num(self, t):
         tok = t.next_token()
         assert tok.type == TOKEN_STRING or tok.type == TOKEN_NAME or tok.type == TOKEN_NUMBER, "Expected ident/string literal/num. Got %s" % tok
         if tok.type == TOKEN_STRING:
-            tok.value = tok.value.strip("'"+'"')
-        return tok    
-        
+            tok.value = tok.value.strip("'" + '"')
+        return tok
+
     def read_string(self, t):
         tok = t.next_token()
         assert tok.type == TOKEN_STRING, "Expected string literal. Got %s" % tok
         if tok.type == TOKEN_STRING:
-            tok.value = tok.value.strip("'"+'"')
+            tok.value = tok.value.strip("'" + '"')
         return tok
-        
+
     def read_equals(self, t):
         tok = t.next_token()
         assert tok.value == '=', "Expected =, Got %s" % tok
-        return tok    
-        
+        return tok
+
     def read_number(self, t):
         tok = t.next_token()
-        assert (tok.type == TOKEN_STRING or tok.type==TOKEN_NUMBER), "Expect Number. Got: %s" % tok
-        return tok  
-        
+        assert (tok.type == TOKEN_STRING or tok.type == TOKEN_NUMBER), "Expect Number. Got: %s" % tok
+        return tok
+
     def read_name2(self, t):
         tok = t.next_token()
         name = tok.value
         self.assert_token(name, TOKEN_NAME, message="Expected identifier. Got: %s" % name)
-        return name    
-        
+        return name
+
     def read_semi(self, t):
         semi = t.next_token(False)
         self.assert_token(semi, TOKEN_OP, value=";", message='Expected ; Got' + str(semi.value))
-        
+
         # read any possible comments after a semi
-        token =  t.readahead()
+        token = t.readahead()
         if token.type == TOKEN_OP and token.value == '//':
-            node = self.read_single_line_comment(t)   
-    
+            node = self.read_single_line_comment(t)
+
     def assert_read(self, t, match_type, expected):
         next = t.next_token()
         self.assert_token(next, match_type, value=expected, message='Expected %s. Got %s' % (expected, next))
-        
+
     def assert_token(self, token, type, value=None, message=None):
         message = "Expected %s @ [%s:%s]" % (value, token.start[0], token.start[1])
-            
+
         pe = ParseError(message)
         pe.line = token.start[0]
         pe.col = token.start[1]
-        
+
         if value is None:
             if token.type != type:
                 raise pe
         else:
             if token.type != type and token.value != value:
                 raise pe
-                
-                
+
+
     def parse_service_stmt(self, t):
         '''
             service_defs = (option	 | method_def  | ';')+ ;
@@ -254,26 +276,26 @@ class ProtoParser(object):
         '''
         node = ServiceNode()
         node.name = self.read_name(t)
-        
+
         self.assert_read(t, TOKEN_OP, '{')
-        
+
         while t.has_next():
-            token =  t.next_token()
+            token = t.next_token()
             if self.check_names(token, ["option"]):
                 node.add_child(self.parse_option_stmt(t))
-                
+
             elif self.check_names(token, ["rpc"]):
                 node.add_child(self.parse_method_def(t))
-                
+
             elif self.check_single_line_comment(token):
                 node.add_child(self.read_single_line_comment(t))
             elif self.check_multiline_line_comment(t, token):
-                self.protonode.add_child(self.read_multi_line_comment(tokenizer))    
+                self.protonode.add_child(self.read_multi_line_comment(tokenizer))
             elif token.type == TOKEN_OP and token.value == '}':
                 break
-                
+
         return node
-        
+
     def parse_method_def(self, t):
         ''' method_options = '{' (option | ';')+ '}' ;
             method_def = 'rpc' ident '(' qualified_name ')' 'returns' '(' qualified_name ')' (method_options | ';') ;
@@ -288,35 +310,36 @@ class ProtoParser(object):
         node.response_type = self.read_name(t)
         self.assert_read(t, TOKEN_OP, ')')
         return node
-        
-        
+
+
     def parse_option_stmt(self, t):
         '''
             option = 'option' optionBody ';';
             optionBody = ident ( '.' ident )* '=' constant;
+            
+            # XXX: what are the parens for: "option (key) = value;" ?
         '''
         node = OptionNode()
-        
-        token =  t.readahead()
-            
+
+        token = t.readahead()
+
         has_open_paren = False
         if token.type == TOKEN_OP and token.value == '(':
-            has_open_paren  = True
-            open_paren =  t.next_token()
-            
-        
+            has_open_paren = True
+            open_paren = t.next_token()
+
         node.name = self.read_name(t)
-        
+
         if has_open_paren:
             self.assert_read(t, TOKEN_RPAR, ')')
-        
+
         self.read_equals(t)
-        node.value = self.read_string_or_id(t).value    
+        node.value = self.read_string_or_id(t).value
         self.read_semi(t)
-        
-        logger.debug('Read option %s',  node)     
+
+        logger.debug('Read option %s', node)
         return node
-   
+
     def parse_optional_stmt(self, t):
         node = OptionalNode()
         node.name = self.read_name(t)
@@ -325,21 +348,21 @@ class ProtoParser(object):
         node.value = self.read_number(t).value
         self.read_semi(t)
         return node
-        
+
     def parse_message_stmt(self, t):
         '''
             message ='message' ident messageBody;
             messageBody ='{' ( field | enum | message | extend | extensions | group | option | ':' )* '}';
         '''
         message_name = self.read_name(t)
-        
+
         self.assert_read(t, TOKEN_OP, '{')
-        
+
         mnode = MessageNode()
         mnode.name = message_name
         logger.debug('Read message node %s', message_name)
         while t.has_next():
-            token =  t.next_token()
+            token = t.next_token()
             if self.check_names(token, ['optional', 'repeated', 'required']):
                 #logger.debug('Read field %s', message_name)
                 # allow parse_field_desc to read the field label
@@ -347,32 +370,32 @@ class ProtoParser(object):
                 node = self.parse_field_desc(t)
                 node.parent = mnode
                 mnode.add_child(node)
-                
+
             elif self.check_single_line_comment(token):
                 node = self.read_single_line_comment(t)
                 mnode.add_child(node)
             elif self.check_multiline_line_comment(t, token):
-                self.protonode.add_child(self.read_multi_line_comment(tokenizer))    
+                self.protonode.add_child(self.read_multi_line_comment(tokenizer))
             elif self.check_names(token, ['extensions']):
                 node = self.parse_extensions_stmt(t)
                 mnode.add_child(node)
-                
+
             elif self.check_names(token, ['enum']):
                 node = self.parse_enum_stmt(t)
                 node.parent = mnode
                 mnode.add_child(node)
-                
+
             elif self.check_names(token, ['message']):
                 node = self.parse_message_stmt(t)
                 node.parent = mnode
                 mnode.add_child(node)
-                
+
             elif token.type == TOKEN_OP and token.value == '}':
                 break
-        
-        logger.debug('End message node %s [%s:%s]', message_name, t.line, t.column)        
+
+        logger.debug('End message node %s [%s:%s]', message_name, t.line, t.column)
         return mnode
-    
+
     def parse_extensions_stmt(self, t):
         ''' 
             # extension numbers must not overlap with field or other extension numbers;
@@ -382,19 +405,19 @@ class ProtoParser(object):
         '''
         ext = ExtensionsNode()
         ext.start = t.next_token().value
-        totoken =  t.next_token()
-        ext.end =  t.next_token().value
+        totoken = t.next_token()
+        ext.end = t.next_token().value
         self.read_semi(t)
         return ext
-        
+
     def parse_syntax_stmt(self, t):
-        node  = SyntaxNode()
+        node = SyntaxNode()
         self.read_equals(t)
         node.value = self.read_string(t).value
         self.read_semi(t)
         logger.debug('Read syntax node %s', node)
         return node
-        
+
     def parse_enum_stmt(self, t):
         '''
             enum ='enum' ident '{' ( option | enumField | ';' )* '}';
@@ -405,13 +428,13 @@ class ProtoParser(object):
         enum.name = t.next_token().value
         self.assert_read(t, TOKEN_OP, '{')
         while t.has_next():
-            token =  t.readahead()
-            
+            token = t.readahead()
+
             if token.type == TOKEN_OP and token.value == '}':
                 t.skip()
                 break
             if token.type == TOKEN_SPACE:
-                
+
                 t.skip(line=1 if token.value == "\n" else 0)
             elif self.check_single_line_comment(token):
                 node = self.read_single_line_comment(t)
@@ -422,26 +445,26 @@ class ProtoParser(object):
                 enum.add_child(self.parse_enumassignment_stmt(t))
             else:
                 raise Exception("Error parsing enum")
-        logger.debug('Read enum %s', enum)       
+        logger.debug('Read enum %s', enum)
         return enum
-        
+
     def parse_enumassignment_stmt(self, t):
         enumassign = EnumAssignmentNode()
-        enumassign.name =  t.next_token().value
+        enumassign.name = t.next_token().value
         self.read_equals(t)
-        enumassign.value = self.read_name(t) 
+        enumassign.value = self.read_name(t)
         self.read_semi(t)
         return enumassign
-        
+
     def parse_extend_stmt(self, t):
         ''' 
             extend ='extend' userType '{' ( field | group | ';' )* '}';
         '''
         extend = ExtendNode()
         extend.name = t.next_token().value
-        token =  t.next_token()
+        token = t.next_token()
         while t.has_next():
-            token =  t.next_token()
+            token = t.next_token()
             if token.type == TOKEN_OP and token.value == '}':
                 break
             elif self.check_single_line_comment(token):
@@ -454,48 +477,55 @@ class ProtoParser(object):
             else:
                 raise Exception("Unexpected token: %s" % token)
         return extend
-        
+
     def parse_import_stmt(self, t):
         ''' import ='import' strLit ';';
         '''
+        # TODO: handle import public strLit;
+        
         import_path = self.read_string_or_id(t)
         semi = t.next_token()
         assert semi.type == TOKEN_OP and semi.value == ';', 'Expected semicolon'
         node = ImportNode()
         node.value = import_path.value
-        logger.debug('Read import node %s', node)  
+        logger.debug('Read import node %s', node)
         return node
-        
+
     def parse_package_stmt(self, t):
         ''' package ='package' ident ( '.' ident )* ';';
         '''
         package_name = t.read_id()
         semi = self.read_semi(t)
-        pnode = PackageNode()
-        pnode.name = package_name
-        logger.debug('Read package node %s', pnode)  
-        return pnode
-        
+        package = PackageDefinition()
+        package.name = package_name
+
+        logger.debug('Read package node %s', package)
+
+        # update name of the the root package node
+        self.root.name = package_name
+
+        return package
+
     def parse_field_desc(self, t):
         '''
             field =label type ident '=' intLit ( '[' fieldOption ( ',' fieldOption )* ']' )? ';';
             # tag number must be 2^29-1 or lower, not 0, and not 19000-19999 (reserved);
             
             fieldOption =optionBody | 'default' '=' constant;
-        '''        
+        '''
         fdn = FieldDescriptorNode()
-        fdn.label =  t.next_token().value
-        fdn.type =   t.read_id()
+        fdn.label = t.next_token().value
+        fdn.type = t.read_id()
         fdn.name = t.next_token().value
-        equals =  t.next_token()
-        assert  equals.type == TOKEN_OP and equals.value == '=', str(equals)
-        number =  t.next_token()
-        assert  number.type == TOKEN_NUMBER
+        equals = t.next_token()
+        assert equals.type == TOKEN_OP and equals.value == '=', "Missing equals. Got %s" % str(equals)
+        number = t.next_token()
+        assert number.type == TOKEN_NUMBER, "Tag should be a number"
         if number.value.startswith('0x'):
             fdn.number = int(number.value, 16)
         else:
             fdn.number = int(number.value, 10)
-            
+
         # group nodes have a braces after with a list of fielddesc inside
         if fdn.type == "group":
             next = t.readahead()
@@ -508,30 +538,30 @@ class ProtoParser(object):
             next = t.readahead()
             if next.value == '[':
                 brace = t.next_token()
-                
+
                 while t.has_next():
                     name, value = self.parse_field_desc_attr(t)
                     if name == 'default':
                         fdn.default = value
                     logger.debug('Parsed field attr %s: %s', name, value)
-                    
+
                     next = t.next_token()
-                    
+
                     if next.value == ']':
                         break
                     elif next.value == ',':
                         logger.debug('read next field')
                         pass
-                
+
             self.read_semi(t)
         logger.debug('Read field %s', fdn)
         return fdn
-        
+
     def parse_group_body(self, t):
         children = []
         while t.has_next():
-            token =  t.next_token()
-            
+            token = t.next_token()
+
             if token.type == TOKEN_NAME and token.value in ['optional', 'repeated', 'required']:
                 # allow parse_field_desc to read the field label
                 t.backtrack()
@@ -541,7 +571,7 @@ class ProtoParser(object):
                 #t.backtrack()
                 break
         return children
-        
+
     def parse_field_desc_attr(self, t):
         name = []
         next = t.next_token()
@@ -555,10 +585,9 @@ class ProtoParser(object):
             name.append(next.value)
         self.read_equals(t)
         value = self.read_string_or_id_or_num(t)
-        
+
         return ''.join(name), value.value
-        
-    
-            
+
+
 #tokens, remainder = scanner.scan(s)
 #print tokens, remainder[0:20]
